#pragma once
#include <math.h>

namespace ur {
const double alpha1 = M_PI_2;
const double alpha4 = M_PI_2;
const double alpha5 = -M_PI_2;
}  // namespace ur
#pragma once

namespace ur3e {
const double d1 = 0.15185;
const double d4 = 0.13105;
const double d5 = 0.08535;
const double d6 = 0.0921;
const double a2 = -0.24355;
const double a3 = -0.2132;
}  // namespace ur3e

namespace ur5e {
const double d1 = 0.1625;
const double d4 = 0.1333;
const double d5 = 0.0997;
const double d6 = 0.0996;
const double a2 = -0.425;
const double a3 = -0.3922;
}  // namespace ur5e

namespace ur10e {
const double d1 = 0.1807;
const double d4 = 0.17415;
const double d5 = 0.11985;
const double d6 = 0.11655;
const double a2 = -0.6127;
const double a3 = -0.57155;
}  // namespace ur10e
